# 
# This one seems to be working pretty well
# 
# Thu Apr 6 00:12:37 EDT 2023
# 
# Now I've attempted to speed things up by using the rfft and irfft commands
# from scipy, per the instructions in
# https://realpython.com/python-scipy-fft/#making-it-faster-with-rfft
# 
# In a lot of ways, this helps make things more sane: I don't have the
# duplication in the audiogram, the frequencies, etc. The sound is pretty good 
# 
# This one uses a callback function, based on wire_callback.py from the
# pyaudio.git example files, at /Users/longa/github/pyaudio.git/examples  
# 
import sys
import time

import pyaudio
import numpy as np
from scipy.fft import rfft
from scipy.fft import irfft
from scipy.interpolate import CubicHermiteSpline

import matplotlib.pyplot as plt

p = pyaudio.PyAudio()
sample_rate=44100
int_size=2**16
chunk_size=2**10
chunk_half=int(chunk_size/2)

# 
# ael:
# I'm using my own audiogram (with some falk info below 250 Hz, and after 8000
# Hz), but I have to adjust it to try to pretend it's in power (rather than
# decibels):  
hz =    np.array([31.25, 62.5, 125.0, 250.0, 500.0, 750.0, 1000.0, 2000.0, 3000.0, 4000.0, 6000.0, 8000.0, 8500,9000,9500,10000])
power = np.array([   10,   10,    20,    25,    40,    50,     50,     55,     60,     65,     70,     70,   60,  30,   0,    0])
# lea

dydx = np.gradient(power,hz)
Audiogram = CubicHermiteSpline(hz, power, dydx)

vMax=110
def tooBig(x):
    return vMax if x > vMax else x
def tooSmall(x): # this used to be set to 0, but it seemed to produce some buzzing.
    return 1 if x < 1 else x

freqz = np.arange(0, chunk_half+1, 1) * sample_rate / chunk_size
# print(freqz)
plt.figure()
plt.plot(hz, power, 'o', label='data')
plt.plot(freqz, [tooSmall(tooBig(Audiogram(x))) for x in freqz])
plt.ylabel('threshold volume')
plt.title('Cubic Spline-Fit Audiogram from Testing')
# plt.show()
# plt.show(block=False)
# plt.ion()
# plt.pause(0.01)

# now we populate the audiogram vector, using the Audiogram function:
audiogram = np.ones(chunk_half+1)
for i in range(0,chunk_half,1):
    audiogram[i] = tooBig(tooSmall(Audiogram(freqz[i])))
#    print(freqz[i],audiogram[i])
#
# Let's plot the Audiogram:
#
plt.figure()
plt.plot(range(0,chunk_half+1,1), audiogram, 'o', label='audiogram')
plt.title('Cubic Spline-Fit Audiogram from Testing')
# plt.show()
# plt.show(block=False)
# plt.pause(0.01)

def callback(in_data, frame_count, time_info, status):
    
    data_in = np.frombuffer(in_data, dtype=np.float32)
    
    # perform FFT:
    fft_data = rfft(data_in)
    
    # get the magnitudes of the positive frequencies:
    fft_data_size = np.abs(fft_data)
    energy_before=np.mean(fft_data_size)
    
    # apply the audiogram, multiplying to reweight the frequencies:
    fft_data_altered = fft_data * audiogram
    
    # now since we've altered the data, we should renormalize, so let's see
    # how much energy our new signal has:
    fft_data_altered_size = np.abs(fft_data_altered)
    energy_after=np.mean(fft_data_altered_size)
    
    # rescale the altered fft_data, so that it has the energy that the
    # signal started with (although distributed differently):
    fft_data_altered = fft_data_altered * energy_before / energy_after
    
    # It's doing what I expect: the mean fft_data is back to what it was at
    # the outset, after having been scaled by the audiogram:
    # fft_data_altered_size = np.abs(fft_data_altered)
    # energy_final=np.mean(fft_data_altered_size)
    # print(energy_before, energy_after, energy_final)
    # lea
    ift=irfft(fft_data_altered)
    
    # here's the output:
    out_data = np.array(ift, dtype='float32').tobytes()
    
    return (out_data, pyaudio.paContinue)

stream = p.open(format=pyaudio.paFloat32, 
                channels=1 if sys.platform == 'darwin' else 2,
                rate=44100,
                input=True,
                output=True,
                frames_per_buffer=chunk_size,
                stream_callback=callback)

start = time.time()
while stream.is_active():
    time.sleep(0.1)

stream.close()
p.terminate()

